from django.db.models import CharField
from django.utils.translation import gettext_lazy as _

from . import validators

__all__ = [
    'CardanoAddressField',
]


class CardanoAddressField(CharField):
    default_validators = [validators.validate_cardano_address]
    description = _("Cardano address")

    def __init__(self, *args, **kwargs):
        # TODO: Determine what the actual max length of a cardano address may be
        kwargs.setdefault('max_length', 200)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # We do not exclude max_length if it matches default as we want to change
        # the default in future.
        return name, path, args, kwargs



import re
import subprocess

from django_cardano.settings import django_cardano_settings as settings

from .exceptions import CardanoError

# Output of 'transaction calculate-min-fee' command is presumed
# to be of the exact form: '<int> Lovelace'
MIN_FEE_RE = re.compile(r'(\d+)\s+Lovelace')

# Output of 'query utxo' command is presumed to yield an ASCII table
# containing rows of the form: <TxHash>    <TxIx>      <Amount>
UTXO_RE = re.compile(r'(\w+)\s+(\d+)\s+(.*)')


class CardanoCLI:
    @classmethod
    def run(cls, command, *args, **kwargs) -> str:
        """
        Invoke the specified cardano-cli command/subcommand
        The *args serve as a series of (arg_name, arg_value) tuples
        The **kwargs behave as singular command arguments.

        :param command: command/subcommand to invoke
        :param args:  Tuples containing optional argument name/value pairs
        :param kwargs: Additional argument name/value pairs
        :return: The cardano-cli command output written to stdout
        """
        process_args = [settings.CLI_PATH] + command.split()

        for arg in args:
            if isinstance(arg, str):
                process_args.append(f'--{arg}')
            elif isinstance(arg, tuple) and len(arg) == 2:
                process_args.append(f'--{arg[0]}')
                process_args.append(arg[1])

        options = dict(kwargs)
        if 'network' in options:
            if options['network'] == 'mainnet':
                process_args.append('--mainnet')
            elif options['network'] == 'testnet':
                process_args += ['--testnet-magic', str(settings.TESTNET_MAGIC)]
            del options['network']

        for option_name, option_value in options.items():
            process_args.append(f'--{option_name}')
            if option_value is not None:
                if isinstance(option_value, list):
                    process_args += option_value
                elif isinstance(option_value, tuple):
                    process_args += list(option_value)
                else:
                    process_args.append(str(option_value))

        subprocess_args = {
            'check': True,
            'capture_output': True,
            'env': {'CARDANO_NODE_SOCKET_PATH': settings.NODE_SOCKET_PATH},
        }

        if command == 'transaction build-raw':
            # Iff the CLI command being invoked is "transaction build-raw",
            # issue the stringified command in shell mode.
            #
            # The reason for this is that  when performing a transaction
            # that involved native tokens, the --tx-out argument(s) contain
            # a space (ex: <addr>+<lovelace>+"<quantity> <asset_id>").
            #
            # For whatever reason, subprocess.run(...) does not permit the use
            # of arguments containing spaces, thus the composed/executed command
            # is not of the intended form and will consequently fail.
            #
            # Enabling shell mode allows the command to be invoked with spaces
            # and quotation characters intact.
            subprocess_args['shell'] = True
            process_args = ' '.join(process_args)

        try:
            completed_process = subprocess.run(process_args, **subprocess_args)
            if completed_process.returncode == 0:
                return completed_process.stdout.decode().strip()
            else:
                error_message = completed_process.stderr.decode().strip()
                raise CardanoError(error_message)
        except (FileNotFoundError, subprocess.CalledProcessError) as e:
            raise CardanoError(source_error=e)



import os

from django.core.files.storage import FileSystemStorage


class CardanoDataStorage(FileSystemStorage):
    def __init__(self, *args, **kwargs):
        location = os.environ.get('CARDANO_APP_DATA_PATH')

        super().__init__(location, *args, **kwargs)



import subprocess
from enum import Enum

class CardanoErrorType(Enum):
    SIGNING_KEY_DECRYPTION_FAILURE = -2
    POLICY_SIGNING_KEY_DECRYPTION_FAILURE = -4


class CardanoError(Exception):
    def __init__(self, reason=None, source_error=None, code=-1):
        self.code = code
        self.reason = reason

        if source_error and not reason:
            self.reason = str(source_error)

        if isinstance(source_error, subprocess.CalledProcessError):
            cmd = source_error.cmd
            if isinstance(cmd, list):
                cmd = ' '.join(source_error.cmd)
            self.cmd = cmd
            self.process_error = source_error
            self.code = source_error.returncode
            self.reason = str(source_error.stderr)


    def __str__(self) -> str:
        return self.reason



import os
import random
import shutil
from pathlib import Path

from django.conf import settings
from django.test import TestCase
from django.utils.text import slugify

from .exceptions import CardanoError
from .models import (
    get_minting_policy_model,
    get_transaction_model,
    get_wallet_model,
)
from .settings import django_cardano_settings
from .util import CardanoUtils

MintingPolicy = get_minting_policy_model()
Transaction = get_transaction_model()
Wallet = get_wallet_model()

DEFAULT_SPENDING_PASSWORD = 'fL;$qR9FZ3?stf-M'
DEFAULT_MINTING_PASSWORD = 'eMgP3AjU&6KRVTrU'

TOKEN_BUNDLE_PARTS = [
    '"1 fe1249f6a018ccc7a620df6226d6b9b9a63555593051b79885dc2e27"',
    '"1 fe1249f6a018ccc7a620df6226d6b9b9a63555593051b79885dc2e28.TestNFT"',
    '"1 fe1249f6a018ccc7a620df6226d6b9b9a63555593051b79885dc2e29.TestNFT2"',
    '"1 fe1249f6a018ccc7a620df6226d6b9b9a63555593051b79885dc2e29.TestNFT3"',
    '"1 fe1249f6a018ccc7a620df6226d6b9b9a63555593051b79885dc2e29.TestNFT4"',
]
DEFAULT_TOKEN_BUNDLE = ' '.join(TOKEN_BUNDLE_PARTS)

def data_path_for_model(instance):
    base_path = Path(django_cardano_settings.APP_DATA_PATH)
    model_name = slugify(instance._meta.verbose_name)
    return base_path / model_name / str(instance.id)


class CardanoUtilTestCase(TestCase):
    def test_query_tip(self):
        tip_info = CardanoUtils.query_tip()

        self.assertIn('block', tip_info)
        self.assertIn('epoch', tip_info)
        self.assertIn('hash', tip_info)
        self.assertIn('slot', tip_info)

    def test_refresh_protocol_parameters(self):
        protocol_parameters = CardanoUtils.refresh_protocol_parameters()
        self.assertIn('minUTxOValue', protocol_parameters)
        self.assertIn('txFeePerByte', protocol_parameters)

    def test_token_bundle_info(self):
        bundle_info = CardanoUtils.token_bundle_info(DEFAULT_TOKEN_BUNDLE)

        self.assertIn('asset_ids', bundle_info)
        self.assertTrue(isinstance(bundle_info['asset_ids'], list))
        self.assertIn('distinct_policy_ids', bundle_info)
        self.assertTrue(isinstance(bundle_info['distinct_policy_ids'], set))
        self.assertIn('distinct_asset_names', bundle_info)
        self.assertTrue(isinstance(bundle_info['distinct_asset_names'], set))
        self.assertIn('tokens', bundle_info)
        self.assertTrue(isinstance(bundle_info['tokens'], dict))

    def test_token_bundle_size(self):
        bundle_size = CardanoUtils.token_bundle_size(DEFAULT_TOKEN_BUNDLE)
        self.assertEqual(bundle_size, 30)

    def test_min_token_dust_value(self):
        min_token_dust_value = CardanoUtils.min_token_dust_value(DEFAULT_TOKEN_BUNDLE)
        print('How to validate this???', min_token_dust_value)


class DjangoCardanoTestCase(TestCase):
    wallet = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        test_data_path = settings.PROJECT_PATH / 'data' / 'test'
        cls.wallet = Wallet.objects.create_from_path(test_data_path)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        # Discard the associated key files
        shutil.rmtree(data_path_for_model(cls.wallet))

    def test_create_wallet(self):
        try:
            wallet = Wallet.objects.create(
                name='Test Wallet',
                password=DEFAULT_SPENDING_PASSWORD
            )

            address_info = CardanoUtils.address_info(wallet.payment_address)
            self.assertEqual(address_info['type'], 'payment')
            self.assertEqual(address_info['encoding'], 'bech32')
            self.assertEqual(address_info['era'], 'shelley')
            self.assertEqual(address_info['address'], wallet.payment_address)

            shutil.rmtree(data_path_for_model(wallet))
        except CardanoError as e:
            print(e)

    def test_get_address_info(self):
        address_info = CardanoUtils.address_info(self.wallet.payment_address)
        self.assertTrue(isinstance(address_info, dict))

    def test_get_utxos(self):
        utxos = self.wallet.utxos
        self.assertTrue(isinstance(utxos, list))

        for utxo in utxos:
            self.assertIn('TxHash', utxo)
            self.assertIn('TxIx', utxo)
            self.assertIn('Tokens', utxo)

    def test_get_wallet_balance(self):
        tokens, _ = self.wallet.balance
        self.assertTrue(isinstance(tokens, dict))

    def test_send_lovelace(self):
        lovelace_requested = 1000000
        to_address = self.wallet.payment_address

        draft_transaction = self.wallet.send_lovelace(
            lovelace_requested,
            to_address=to_address,
        )
        draft_tx_fee = draft_transaction.calculate_min_fee()

        self.assertTrue(isinstance(draft_transaction, Transaction))
        self.assertTrue(isinstance(draft_tx_fee, int))
        self.assertTrue(draft_transaction._state.adding)

        intermediate_file_path = draft_transaction.intermediate_file_path
        self.assertTrue(intermediate_file_path.exists())

        draft_transaction.delete()

        # Ensure that the intermediate files were deleted
        self.assertFalse(intermediate_file_path.exists())

        transaction = self.wallet.send_lovelace(
            lovelace_requested,
            to_address=to_address,
            password=DEFAULT_SPENDING_PASSWORD,
        )
        self.assertFalse(transaction._state.adding)
        self.assertFalse(transaction.intermediate_file_path.exists())

    def test_send_tokens(self):
        self.wallet.send_tokens(
            'd491fdc194c0d988459ce05a65c8a52259433e84d7162765570aa581.MMTestTokenTwo',
            1,
            to_address=self.wallet.payment_address,
        )

    def test_consolidate_utxos(self):
        self.wallet.consolidate_utxos(password=DEFAULT_SPENDING_PASSWORD)

    def test_partition_lovelace(self):
        min_value = 1000000
        max_value = 5000000
        values = [random.randint(min_value, max_value) for i in range(0, 10)]
        transaction = self.wallet.partition_lovelace(
            values=values,
            password=DEFAULT_SPENDING_PASSWORD
        )
        print(transaction.tx_id)

    def test_create_minting_policy(self):
        tip = CardanoUtils.query_tip()
        invalid_hereafter = tip['slot'] + django_cardano_settings.DEFAULT_TRANSACTION_TTL

        minting_policy = MintingPolicy.objects.create(
            password=DEFAULT_SPENDING_PASSWORD,
            invalid_hereafter=invalid_hereafter,
        )
        policy_script_path = Path(minting_policy.script.path)
        self.assertTrue(policy_script_path.exists())

        # Scrap the generated policy script and associated keys
        shutil.rmtree(data_path_for_model(minting_policy))

    def test_mint_nft(self):
        current_slot = int(CardanoUtils.query_tip()['slot'])
        invalid_hereafter = current_slot + django_cardano_settings.DEFAULT_TRANSACTION_TTL

        policy = MintingPolicy.objects.create(
            password=DEFAULT_MINTING_PASSWORD,
            invalid_hereafter=invalid_hereafter,
        )
        asset_name = 'TestNFT'

        tx_metadata = {
            "721": {
                policy.policy_id: {
                    asset_name: {
                        'name': 'django-cardano Test NFT',
                        'description': 'The Cardano Logo (in SVG format)',
                        'image': 'ipfs://QmS5gynkFMNFTnqPGgADxbvjutYmQK4Qh4iWwkSq4BhcmJ',

                    }
                }
            }
        }

        transaction = self.wallet.mint_tokens(
            policy=policy,
            quantity=1,
            to_address=self.wallet.payment_address,
            spending_password=DEFAULT_SPENDING_PASSWORD,
            minting_password=DEFAULT_MINTING_PASSWORD,
            asset_name=asset_name,
            metadata=tx_metadata,
        )

        # Scrap the generated policy script and associated keys
        shutil.rmtree(data_path_for_model(policy))



import os
import re
from pathlib import Path

from django_cardano.settings import django_cardano_settings as settings

ALPHANUMERIC_RE = re.compile(r'[^a-zA-Z0-9]')


def filter_utxos(utxos, include=None, exclude=None) -> list:
    filtered_utxos = []
    lovelace_unit = settings.LOVELACE_UNIT

    for utxo in utxos:
        tokens = utxo['Tokens']
        asset_types = tuple(tokens.keys())

        if include == lovelace_unit:
            if len(asset_types) == 1:
                # Implicitly, if there is only one asset type in this UTxO
                # it MUST be lovelace. Cardano does not (yet) support the
                # notion of a UTxO without any amount of lovelace.
                filtered_utxos.append(utxo)
        elif include in asset_types:
            filtered_utxos.append(utxo)

        if exclude == lovelace_unit:
            if len(asset_types) > 1:
                # See logical explanation above.
                filtered_utxos.append(utxo)
            elif exclude not in asset_types:
                filtered_utxos.append(utxo)

    return filtered_utxos


def sort_utxos(utxos, type=settings.LOVELACE_UNIT, order='desc') -> list:
    if order == 'desc':
        return sorted(utxos, key=lambda v: v['Tokens'][type], reverse=True)
    else:
        return sorted(utxos, key=lambda v: v['Tokens'][type])


def clean_token_asset_name(asset_name: str) -> str:
    """
    :param asset_name: The asset_name segment of a Cardano native token

    Cardano native assets are identified by the concatenation of
    their policy ID and an optional name:
    <asset_id> = <policy_id>.<asset_name>

    The asset name is restricted to alphanumeric characters, so
    use this shortcut to exclude invalid characters.
    """
    return ALPHANUMERIC_RE.sub('', asset_name)



import json
import math
import os
import re
from collections import defaultdict
from datetime import datetime, timezone
from pathlib import Path

from .cli import CardanoCLI
from .settings import django_cardano_settings as settings

TOKEN_BUNDLE_RE = re.compile(r'(?:\".*?\"|\S)+')


def quot(a: int, b: int) -> int:
    return math.floor(a / b)


def roundup_bytes_to_words(b: int) -> int:
    return quot(b + 7, 8)


class CardanoUtils:
    protocol_parameters_path = Path(settings.APP_DATA_PATH, 'protocol.json')

    @classmethod
    def refresh_protocol_parameters(cls, force=False) -> dict:
        if not os.path.exists(settings.APP_DATA_PATH):
            os.makedirs(settings.APP_DATA_PATH, 0o755)

        load = True
        if cls.protocol_parameters_path.exists():
            if force or not settings.PROTOCOL_TTL:
                load = True
            else:
                file_stats = cls.protocol_parameters_path.stat()
                date_modified = datetime.fromtimestamp(file_stats.st_mtime, tz=timezone.utc)
                now = datetime.now(tz=timezone.utc)
                file_age = now - date_modified
                load = True if file_age.seconds > settings.PROTOCOL_TTL else False

        if load:
            CardanoCLI.run('query protocol-parameters', **{
                'network': settings.NETWORK,
                'out-file': cls.protocol_parameters_path,
            })

        with open(cls.protocol_parameters_path, 'r') as protocol_parameters_file:
            protocol_parameters = json.load(protocol_parameters_file)

        return protocol_parameters

    @classmethod
    def query_tip(cls) -> dict:
        response = CardanoCLI.run('query tip', network=settings.NETWORK)
        return json.loads(response)

    @classmethod
    def address_info(cls, address):
        response = CardanoCLI.run('address info', address=address)
        return json.loads(response)

    @classmethod
    def tx_info(cls, tx_file):
        return CardanoCLI.run('transaction view', **{
            'tx-file': tx_file
        })

    @classmethod
    def token_bundle_info(cls, token_bundle: str) -> dict:
        """
        :param token_bundle: A token bundle
        :return: A dictionary containing three sets:
         - asset_ids: List of asset IDs in the bundle
         - policy_ids: Distinct set of policy IDs
         - asset_names: Distinct set of asset names
        """
        asset_ids = []
        tokens = defaultdict(int)
        distinct_policy_ids = set()
        distinct_asset_names = set()

        for bundle_entry in TOKEN_BUNDLE_RE.findall(token_bundle):
            bundle_entry = bundle_entry.strip('"')
            token_count, asset_id = bundle_entry.split(' ')
            tokens[asset_id] += int(token_count)
            asset_ids.append(asset_id)

            try:
                policy_id, asset_name = asset_id.split('.')
                distinct_policy_ids.add(policy_id)
                distinct_asset_names.add(asset_name)
            except ValueError:
                distinct_policy_ids.add(asset_id)

        return {
            'asset_ids': asset_ids,
            'distinct_policy_ids': distinct_policy_ids,
            'distinct_asset_names': distinct_asset_names,
            'tokens': tokens,
        }

    @classmethod
    def token_bundle_size(cls, token_bundle: str) -> int:
        """
        :param token_bundle: Full-string representation of token bundle
        :return: Size of the token bundle B in 8-byte long words

        num_assets: the number of distinct AssetIDs in token_bundle
        num_policy_ids: the number of distinct PolicyIDs in token_bundle
        sum_assetname_lengths: the sum of the length of the ByteStrings representing distinct asset names
        pid_size: the length of the hash of a policy (ie. the length of the PolicyID).
        bundle_size is the size of the token bundle B in 8-byte long words:
            ex: bundle_size = 6 + roundupBytesToWords(((numAssets B) * 12) +
                (sumAssetNameLengths B) + ((numPids B) * pid_size))

        TODO: Find an explanation for the '6' and '12' numbers in the above formula
        See: https://cardano-ledger.readthedocs.io/en/latest/explanations/min-utxo.html
        """
        bundle_info = cls.token_bundle_info(token_bundle)

        asset_ids = bundle_info['asset_ids']
        asset_names = bundle_info['distinct_asset_names']
        policy_ids = bundle_info['distinct_policy_ids']
        asset_name_lengths = sum([len(asset_name) for asset_name in asset_names])

        byte_count = (len(asset_ids) * 12) + asset_name_lengths
        for policy_id in policy_ids:
            # Note that policy IDs are represented as hexadecimal byte strings,
            # and one byte will contain two hex values, hence the division by 2.
            # Ex: 0xAF = 10101111 = one byte
            byte_count += math.ceil(len(policy_id) / 2)

        bundle_size = 6 + roundup_bytes_to_words(byte_count)
        return bundle_size

    @classmethod
    def min_token_dust_value(cls, token_bundle: str) -> int:
        """
        See: https://cardano-ledger.readthedocs.io/en/latest/explanations/min-utxo.html

        num_assets: the number of distinct AssetIDs in token_bundle
        num_policy_ids: the number of distinct PolicyIDs in token_bundle
        sum_assetname_lengths: the sum of the length of the ByteStrings representing distinct asset names
        pid_size: the length of the hash of a policy (ie. the length of the PolicyID).
        bundle_size is the size of the token bundle B in 8-byte long words:
            ex: bundle_size = 6 + roundupBytesToWords(((numAssets B) * 12) +
                (sumAssetNameLengths B) + ((numPids B) * pid_size))

        :return: Amount of lovelace (a.k.a. "dust") to accompany a UTxO containing non-ADA tokens
        """
        protocol_parameters = cls.refresh_protocol_parameters()
        min_utxo_value = protocol_parameters['minUTxOValue']
        utxo_entry_size_without_val = settings.UTXO_ENTRY_SIZE_WITHOUT_VAL
        ada_only_utxo_size = utxo_entry_size_without_val + settings.COIN_SIZE

        bundle_size = cls.token_bundle_size(token_bundle)
        return max(
            min_utxo_value,
            quot(min_utxo_value, ada_only_utxo_size) * (utxo_entry_size_without_val + bundle_size)
        )



from django.utils.deconstruct import deconstructible
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from django_cardano.exceptions import CardanoError
from django_cardano.util import CardanoUtils


@deconstructible
class CardanoAddressValidator:
    message = _('Enter a valid cardano address.')
    code = 'invalid'

    def __init__(self, message=None, code=None):
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        try:
            CardanoUtils.address_info(value)
        except CardanoError:
            raise ValidationError(self.message, code=self.code)

    def __eq__(self, other):
        return (
            isinstance(other, CardanoAddressValidator) and
            (self.message == other.message) and
            (self.code == other.code)
        )


validate_cardano_address = CardanoAddressValidator()



import io
import json
import pyAesCrypt
import tempfile
import uuid

from collections import defaultdict
from pathlib import Path
from typing import Optional

from django.db import models
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile
from django.utils.text import slugify

from .cli import (
    CardanoCLI,
    MIN_FEE_RE,
    UTXO_RE,
)

from .fields import CardanoAddressField
from .exceptions import CardanoError, CardanoErrorType
from .util import CardanoUtils

from .shortcuts import (
    filter_utxos,
    sort_utxos,
)
from .storage import CardanoDataStorage

from django_cardano.settings import (
    django_cardano_settings as cardano_settings
)

# Size of buffer supplied used for encryption of signing keys
ENCRYPTION_BUFFER_SIZE = 4 * 1024

lovelace_unit = cardano_settings.LOVELACE_UNIT


# ---------------------------------------------------------------------------------
def get_extensible_model(setting_name):
    model_name = getattr(settings, setting_name)
    try:
        return django_apps.get_model(model_name, require_ready=True)
    except ValueError:
        raise ImproperlyConfigured(f"{setting_name} must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            f"{setting_name} refers to model '{model_name}' that has not been installed"
        )


def get_minting_policy_model():
    """
    Return the MintingPolicy model that is active in this project.
    """
    return get_extensible_model('DJANGO_CARDANO_MINTING_POLICY_MODEL')


def get_transaction_model():
    """
    Return the Transaction model that is active in this project.
    """
    return get_extensible_model('DJANGO_CARDANO_TRANSACTION_MODEL')


def get_wallet_model():
    """
    Return the Wallet model that is active in this project.
    """
    return get_extensible_model('DJANGO_CARDANO_WALLET_MODEL')


# ------------------------------------------------------------------------------
def file_upload_path(instance, filename):
    model_name = slugify(instance._meta.verbose_name)
    return Path(model_name, str(instance.id), filename)


class MintingPolicyManager(models.Manager):
    def create(self, password, invalid_before=None, invalid_hereafter=None, **kwargs):
        policy = self.model(**kwargs)

        with tempfile.TemporaryDirectory() as tmpdirname:
            intermediate_file_path = Path(tmpdirname)

            # 1. Create signing/verification keys for the minting policy
            signing_key_path = intermediate_file_path / 'signing.key'
            verification_key_path = intermediate_file_path / 'verification.key'
            CardanoCLI.run('address key-gen', **{
                'signing-key-file': signing_key_path,
                'verification-key-file': verification_key_path,
            })

            # 2. Encrypt the generated key files and attach them to the Policy record
            for field_name, file_path in {
                'signing_key': signing_key_path,
                'verification_key': verification_key_path,
            }.items():
                with open(file_path, 'rb') as fp:
                    filename = file_path.name
                    f_ciph = io.BytesIO()
                    pyAesCrypt.encryptStream(io.BytesIO(fp.read()), f_ciph, password, ENCRYPTION_BUFFER_SIZE)
                    file_field = getattr(policy, field_name)
                    file_field.save(f'{filename}.aes', f_ciph, save=False)

            policy_key_hash = CardanoCLI.run('address key-hash', **{
                'payment-verification-key-file': verification_key_path,
            })

            # 3. Construct the policy script and attach to the Policy record
            scripts = [{
                'keyHash': policy_key_hash,
                'type': 'sig',
            }]
            if invalid_before:
                scripts.append({
                    'type': 'after',
                    'slot': invalid_before,
                })
            if invalid_hereafter:
                scripts.append({
                    'type': 'before',
                    'slot': invalid_hereafter,
                })
            policy_script_data = json.dumps({
                'type': 'all',
                'scripts': scripts
            })
            with ContentFile(policy_script_data) as file_content:
                policy.script.save('policy.script.json', file_content, save=False)

            # 4. Determine the policy ID (i.e. compute hash of the policy script)
            policy.policy_id = CardanoCLI.run('transaction policyid', **{
                'script-file': policy.script.path
            })

        policy.save(force_insert=True, using=self.db)
        return policy


class AbstractMintingPolicy(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    policy_id = models.CharField(max_length=64)

    name = models.CharField(max_length=50, blank=True)

    script = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )
    signing_key = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )
    verification_key = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )

    objects = MintingPolicyManager()

    class Meta:
        abstract = True

    def __str__(self):
        return self.name if self.name else self.policy_id

    @property
    def script_data(self):
        if not self.script:
            return None

        with open(self.script.path, 'r') as script_file_path:
            return json.load(script_file_path)


class MintingPolicy(AbstractMintingPolicy):
    class Meta:
        swappable = 'DJANGO_CARDANO_MINTING_POLICY_MODEL'
        verbose_name_plural = 'Minting Policies'


# ---------------------------------------------------------------------------------
class TransactionTypes(models.IntegerChoices):
    LOVELACE_TRANSFER = 1
    TOKEN_TRANSFER = 2
    TOKEN_MINT = 3
    TOKEN_CONSOLIDATION = 4
    LOVELACE_PARTITION = 5


class AbstractTransaction(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    tx_id = models.CharField(max_length=64, blank=True, null=True)
    tx_file = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )
    tx_type = models.PositiveSmallIntegerField(choices=TransactionTypes.choices)

    inputs = models.JSONField(default=list)
    outputs = models.JSONField(default=list)
    metadata = models.JSONField(blank=True, null=True)

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.temp_directory = tempfile.TemporaryDirectory()
        self.minting_policy = None
        self.minting_password = None

    def __del__(self):
        self.temp_directory.cleanup()

    def __str__(self):
        return str(self.id)

    @property
    def intermediate_file_path(self) -> Path:
        return Path(self.temp_directory.name)

    @property
    def metadata_file_path(self) -> Path:
        return self.intermediate_file_path / 'metadata.json'

    @property
    def draft_tx_file_path(self) -> Path:
        return self.intermediate_file_path / 'transaction.draft'

    @property
    def tx_info(self) -> Optional[dict]:
        return CardanoUtils.tx_info(self.tx_file.path) if self.tx_file else None

    @property
    def tx_args(self) -> list:
        return self.inputs + self.outputs

    def delete(self, using=None, keep_parents=False):
        # Destroy all intermediate files upon deletion
        self.temp_directory.cleanup()

        return super().delete(using, keep_parents)

    def generate_draft(self, **kwargs):
        cmd_kwargs = {
            **kwargs,
            'fee': 0,
            'out-file': self.draft_tx_file_path,
        }
        if self.metadata:
            with open(self.metadata_file_path, 'w') as metadata_file:
                json.dump(self.metadata, metadata_file)
            cmd_kwargs.update({
                'json-metadata-no-schema': None,
                'metadata-json-file': self.metadata_file_path,
            })

        CardanoCLI.run('transaction build-raw', *self.tx_args, **cmd_kwargs)

    def calculate_min_fee(self) -> int:
        tx_body_file_path = Path(self.draft_tx_file_path)
        if not tx_body_file_path.exists():
            raise CardanoError('Unable to calculate minimum fee; require transaction body file.')

        CardanoUtils.refresh_protocol_parameters()

        raw_response = CardanoCLI.run('transaction calculate-min-fee', **{
            'tx-body-file': tx_body_file_path,
            'tx-in-count': len(self.inputs),
            'tx-out-count': len(self.outputs),
            'witness-count': 2 if self.minting_policy else 1,
            'byron-witness-count': 0,
            'protocol-params-file': CardanoUtils.protocol_parameters_path,
            'network': cardano_settings.NETWORK,
        })
        match = MIN_FEE_RE.match(raw_response)
        return int(match[1])

    def submit(self, wallet, fee, password, invalid_hereafter=None, **tx_kwargs):
        raw_tx_file_path = self.intermediate_file_path / 'transaction.raw'
        signed_tx_file_path = self.intermediate_file_path / 'transaction.signed'
        signing_key_file_path = self.intermediate_file_path / 'signing.key'
        policy_signing_key_file_path = self.intermediate_file_path / 'policy-signing.key'

        # Determine the TTL (time to Live) for the transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#determine-the-ttl-time-to-live-for-the-transaction
        if not invalid_hereafter:
            current_slot = int(CardanoUtils.query_tip()['slot'])
            invalid_hereafter = current_slot + cardano_settings.DEFAULT_TRANSACTION_TTL

        cmd_kwargs = {
            **tx_kwargs,
            'fee': fee,
            'invalid-hereafter': invalid_hereafter,
            'out-file': raw_tx_file_path,
        }

        if self.metadata:
            cmd_kwargs.update({
                'json-metadata-no-schema': None,
                'metadata-json-file': self.metadata_file_path,
            })

        CardanoCLI.run('transaction build-raw', *self.tx_args, **cmd_kwargs)

        # Sign the transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#sign-the-transaction
        signing_args = []
        signing_kwargs = {
            'signing-key-file': signing_key_file_path,
            'tx-body-file': raw_tx_file_path,
            'out-file': signed_tx_file_path,
            'network': cardano_settings.NETWORK
        }

        # Decrypt this wallet's signing key and store it as a temporary file
        try:
            pyAesCrypt.decryptFile(
                wallet.payment_signing_key.path,
                signing_key_file_path,
                password,
                ENCRYPTION_BUFFER_SIZE
            )
        except ValueError as e:
            raise CardanoError(
                source_error=e,
                code=CardanoErrorType.SIGNING_KEY_DECRYPTION_FAILURE
            )

        if self.minting_policy and self.minting_password:
            signing_kwargs['script-file'] = self.minting_policy.script.path

            # Decrypt the policy signing key and store it as a temporary file
            try:
                pyAesCrypt.decryptFile(
                    self.minting_policy.signing_key.path,
                    policy_signing_key_file_path,
                    self.minting_password,
                    ENCRYPTION_BUFFER_SIZE
                )
                signing_args.append(('signing-key-file', policy_signing_key_file_path))
            except ValueError as e:
                raise CardanoError(
                    source_error=e,
                    code=CardanoErrorType.POLICY_SIGNING_KEY_DECRYPTION_FAILURE,
                )

        # Sign the transaction
        CardanoCLI.run('transaction sign', *signing_args, **signing_kwargs)

        self.tx_id = CardanoCLI.run('transaction txid', **{
            'tx-file': signed_tx_file_path
        })

        # Submit the transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#submit-the-transaction
        CardanoCLI.run('transaction submit', **{
            'tx-file': signed_tx_file_path,
            'network': cardano_settings.NETWORK
        })

        with open(signed_tx_file_path, 'rb') as signed_tx_file:
            with ContentFile(signed_tx_file.read()) as file_content:
                self.tx_file.save(signed_tx_file_path.name, file_content, save=False)

        # Clean up intermediate files
        self.temp_directory.cleanup()


class Transaction(AbstractTransaction):
    class Meta:
        swappable = 'DJANGO_CARDANO_TRANSACTION_MODEL'


# ------------------------------------------------------------------------------
class WalletManager(models.Manager):
    use_in_migrations = True

    def create_from_path(self, path, **kwargs):
        wallet = self.model(**kwargs)

        with open(path / 'payment.addr', 'r') as payment_address_file:
            wallet.payment_address = payment_address_file.read()
        with open(path / 'staking.addr', 'r') as staking_address_file:
            wallet.stake_address = staking_address_file.read()

        # Encrypt the test key files and attach them to the wallet
        for field_name, file_path in {
            'payment_signing_key': path / 'signing.key.aes',
            'payment_verification_key': path / 'verification.key.aes',
            'stake_signing_key': path / 'stake-signing.key.aes',
            'stake_verification_key': path / 'stake-verification.key.aes',
        }.items():
            with open(file_path, 'rb') as fp:
                with ContentFile(fp.read()) as file_content:
                    field = getattr(wallet, field_name)
                    field.save(file_path.name, file_content, save=False)

        wallet.full_clean()
        wallet.save(force_insert=True, using=self.db)

        return wallet

    def create(self, password, **kwargs):
        wallet = self.model(**kwargs)

        with tempfile.TemporaryDirectory() as tmp_path:
            intermediate_file_path = Path(tmp_path)

            # Generate the payment signing & verification keys
            signing_key_path = intermediate_file_path / 'signing.key'
            verification_key_path = intermediate_file_path / 'verification.key'
            CardanoCLI.run('address key-gen', **{
                'signing-key-file': signing_key_path,
                'verification-key-file': verification_key_path,
            })

            # Generate the stake signing & verification keys
            stake_signing_key_path = intermediate_file_path / 'stake-signing.key'
            stake_verification_key_path = intermediate_file_path / 'stake-verification.key'
            CardanoCLI.run('stake-address key-gen', **{
                'signing-key-file': stake_signing_key_path,
                'verification-key-file': stake_verification_key_path,
            })

            # Create the payment address.
            wallet.payment_address = CardanoCLI.run('address build', **{
                'payment-verification-key-file': verification_key_path,
                'stake-verification-key-file': stake_verification_key_path,
                'network': cardano_settings.NETWORK,
            })

            # Create the staking address.
            wallet.stake_address = CardanoCLI.run('stake-address build', **{
                'stake-verification-key-file': stake_verification_key_path,
                'network': cardano_settings.NETWORK,
            })

            # Encrypt the generated key files and attach them to the wallet
            for field_name, file_path in {
                'payment_signing_key': signing_key_path,
                'payment_verification_key': verification_key_path,
                'stake_signing_key': stake_signing_key_path,
                'stake_verification_key': stake_verification_key_path,
            }.items():
                with open(file_path, 'rb') as fp:
                    filename = file_path.name
                    f_ciph = io.BytesIO()
                    pyAesCrypt.encryptStream(io.BytesIO(fp.read()), f_ciph, password, ENCRYPTION_BUFFER_SIZE)
                    file_field = getattr(wallet, field_name)
                    file_field.save(f'{filename}.aes', f_ciph, save=False)

        wallet.save(force_insert=True, using=self.db)
        return wallet


class AbstractWallet(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    name = models.CharField(max_length=30, blank=True)

    payment_address = CardanoAddressField()
    payment_signing_key = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )
    payment_verification_key = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )

    stake_address = CardanoAddressField()
    stake_signing_key = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )
    stake_verification_key = models.FileField(
        max_length=200,
        upload_to=file_upload_path,
        storage=CardanoDataStorage
    )

    objects = WalletManager()

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    @property
    def utxos(self) -> list:
        utxos = []

        response = CardanoCLI.run(
            'query utxo',
            address=self.payment_address,
            network=cardano_settings.NETWORK
        )

        lines = response.split('\n')
        for line in lines[2:]:
            match = UTXO_RE.match(line)
            utxo_info = {
                'TxHash': match[1],
                'TxIx': match[2],
                'Tokens': {},
            }

            tokens = match[3].split('+')
            for token in tokens:
                token_info = token.split()
                asset_count = int(token_info[0])
                asset_type = token_info[1]
                utxo_info['Tokens'][asset_type] = asset_count

            utxos.append(utxo_info)

        return utxos

    @property
    def lovelace_utxos(self) -> list:
        return filter_utxos(self.utxos, include=lovelace_unit)

    @property
    def token_utxos(self) -> list:
        return filter_utxos(self.utxos, exclude=lovelace_unit)

    @property
    def balance(self) -> tuple:
        utxos = self.utxos
        all_tokens = defaultdict(int)

        for utxo in utxos:
            utxo_tokens = utxo['Tokens']
            for token_id, token_count in utxo_tokens.items():
                all_tokens[token_id] += token_count

        return all_tokens, utxos

    def send_lovelace(self, quantity, to_address, password=None) -> AbstractTransaction:
        # The protocol's declared txFeeFixed will give us a fair estimate
        # of how much the fee for this transaction will be.
        protocol_parameters = CardanoUtils.refresh_protocol_parameters()
        estimated_tx_fee = protocol_parameters.get('txFeeFixed')

        transaction_class = get_transaction_model()
        transaction = transaction_class(tx_type=TransactionTypes.LOVELACE_TRANSFER)

        # Get the transaction hash and index of the UTxO(s) to spend
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#get-the-transaction-hash-and-index-of-the-utxo-to-spend
        sorted_lovelace_utxos = sort_utxos(self.lovelace_utxos)

        total_lovelace_being_sent = 0
        for utxo in sorted_lovelace_utxos:
            tx_hash = utxo['TxHash']
            tx_index = utxo['TxIx']
            transaction.inputs.append(('tx-in', f'{tx_hash}#{tx_index}'))
            total_lovelace_being_sent += utxo['Tokens'][lovelace_unit]

            # Validate whether the included UTxOs are sufficient to cover
            # the lovelace being transferred, including the estimated tx_fee.
            if total_lovelace_being_sent >= quantity + estimated_tx_fee:
                break

        # There will ALWAYS be exactly two output transactions:
        #   - The funds being sent to the recipient
        #   - The "change" being returned to the sender
        transaction.outputs = [
            ('tx-out', f'{to_address}+{quantity}'),
            ('tx-out', f'{self.payment_address}+{total_lovelace_being_sent}'),
        ]

        # Draft the transaction:
        # Produce a draft transaction in order to determine the fees required to perform the actual transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#draft-the-transaction
        transaction.generate_draft()

        # If a password was given, this implies the intention to commit the
        # transaction to the blockchain (vs. performing a dry-run)
        if password:
            # Calculate the fee
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-fee
            tx_fee = transaction.calculate_min_fee()

            # Calculate the change to return the payment address
            # (minus transaction fee) and update that output respectively
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-change-to-send-back-to-payment-addr
            lovelace_to_return = total_lovelace_being_sent - quantity - tx_fee
            transaction.outputs[-1] = ('tx-out', f'{self.payment_address}+{lovelace_to_return}')

            # Let successful transactions be persisted to the database
            transaction.submit(wallet=self, fee=tx_fee, password=password)
            transaction.save()

        return transaction

    def send_tokens(self, asset_id, quantity, to_address, password=None) -> AbstractTransaction:
        utxos = self.utxos
        sorted_lovelace_utxos = sort_utxos(self.lovelace_utxos)
        token_utxos = sort_utxos(filter_utxos(utxos, include=asset_id), type=asset_id)

        if not sorted_lovelace_utxos:
            # Let there be be at least one UTxO containing purely ADA.
            # This will be used to pay for the transaction.
            raise CardanoError('Insufficient ADA funds to complete transaction')

        transaction_model_class = get_transaction_model()
        transaction = transaction_model_class(tx_type=TransactionTypes.TOKEN_TRANSFER)

        # ASSUMPTION: The largest ADA UTxO shall contain sufficient ADA
        # to pay for the transaction (including fees)
        lovelace_utxo = sorted_lovelace_utxos[0]
        total_lovelace_being_sent = lovelace_utxo['Tokens'][lovelace_unit]
        transaction.inputs = [('tx-in', '{}#{}'.format(lovelace_utxo['TxHash'], lovelace_utxo['TxIx']))]

        # The set of transaction inputs shall be comprised of as many token UTxOs
        # as are required to accommodate the tokens_requested
        total_tokens_being_sent = 0
        for utxo in token_utxos:
            tx_hash = utxo['TxHash']
            tx_index = utxo['TxIx']
            transaction.inputs.append(('tx-in', f'{tx_hash}#{tx_index}'))
            total_tokens_being_sent += utxo['Tokens'][asset_id]

            # Accumulate the total amount of lovelace being sent
            total_lovelace_being_sent += utxo['Tokens'][lovelace_unit]

            if total_tokens_being_sent >= quantity:
                break

        if total_tokens_being_sent < quantity:
            raise CardanoError(f'Insufficient tokens. Requested: {quantity}, Available: {total_tokens_being_sent}')

        lovelace_to_return = total_lovelace_being_sent

        # Let the first transaction output represent the tokens being sent to the recipient
        token_bundle = f'"{quantity} {asset_id}"'
        token_dust = CardanoUtils.min_token_dust_value(token_bundle)
        transaction.outputs = [('tx-out', f'{to_address}+{token_dust}+{token_bundle}')]
        lovelace_to_return -= token_dust

        # If there are more tokens in this wallet than are being sent, return the rest to the sender
        tokens_to_return = total_tokens_being_sent - quantity
        if tokens_to_return > 0:
            token_bundle = f'"{tokens_to_return} {asset_id}"'
            token_dust = CardanoUtils.min_token_dust_value(token_bundle)
            transaction.outputs.append(('tx-out', f'{self.payment_address}+{token_dust}+{token_bundle}'))
            lovelace_to_return -= token_dust

        # The last output represents the lovelace being returned to the payment wallet
        transaction.outputs.append(('tx-out', f'{self.payment_address}+{lovelace_to_return}'))

        # Draft the transaction:
        # Produce a draft transaction in order to determine the fees required to perform the actual transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#draft-the-transaction
        transaction.generate_draft()

        if password:
            # Calculate the fee
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-fee
            tx_fee = transaction.calculate_min_fee()

            # Calculate the change to return the payment address
            # (minus transaction fee) and update that output respectively
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-change-to-send-back-to-payment-addr
            transaction.outputs[-1] = ('tx-out', f'{self.payment_address}+{lovelace_to_return - tx_fee}')

            transaction.submit(wallet=self, fee=tx_fee, password=password)

            # Let successful transactions be persisted to the database
            transaction.save()

        return transaction

    def consolidate_utxos(self, password=None) -> AbstractTransaction:
        all_tokens, utxos = self.balance

        transaction_model_class = get_transaction_model()
        transaction = transaction_model_class(tx_type=TransactionTypes.TOKEN_CONSOLIDATION)

        # Traverse the set of utxos at the given wallet's payment address,
        # accumulating the total count of each type of token.
        for utxo in utxos:
            tx_hash = utxo['TxHash']
            tx_index = utxo['TxIx']
            transaction.inputs.append(('tx-in', f'{tx_hash}#{tx_index}'))

        remaining_lovelace = all_tokens[lovelace_unit]
        del all_tokens[lovelace_unit]

        for asset_id, asset_count in all_tokens.items():
            # HACK!! The amount of ADA accompanying a token needs to be computed
            # with respect to that token's properties
            token_bundle = f'"{asset_count} {asset_id}"'
            token_dust = CardanoUtils.min_token_dust_value(token_bundle)
            transaction.outputs.append(('tx-out', f'{self.payment_address}+{token_dust}+{token_bundle}'))
            remaining_lovelace -= token_dust

        # This output represents the remaining ADA.
        # It must be included in draft transaction in order to accurately compute the
        # minimum transaction fee. After the minimum fee has been calculated,
        # this output will be replaced by one that accounts for that fee.
        transaction.outputs.append(('tx-out', f'{self.payment_address}+{remaining_lovelace}'))

        # Draft the transaction:
        # Produce a draft transaction in order to determine the fees required to perform the actual transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#draft-the-transaction
        transaction.generate_draft()

        if password:
            # Calculate the fee
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-fee
            tx_fee = transaction.calculate_min_fee()

            # Calculate the change to return the payment address
            # (minus transaction fee) and update that output respectively
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-change-to-send-back-to-payment-addr
            transaction.outputs[-1] = ('tx-out', f'{self.payment_address}+{remaining_lovelace - tx_fee}')

            transaction.submit(wallet=self, fee=tx_fee, password=password)

            # Let successful transactions be persisted to the database
            transaction.save()

        return transaction

    def partition_lovelace(self, values: list, password=None) -> AbstractTransaction:
        """
        Convert existing lovelace UTxOs into those of the given values.
        (Useful for testing batch creation of NFTs.)
        :param password:
        :param values:
        :return:
        """
        transaction_model_class = get_transaction_model()
        transaction = transaction_model_class(tx_type=TransactionTypes.LOVELACE_PARTITION)

        surplus_lovelace = 0
        for utxo in self.lovelace_utxos:
            tx_hash = utxo['TxHash']
            tx_index = utxo['TxIx']
            transaction.inputs.append(('tx-in', f'{tx_hash}#{tx_index}'))
            surplus_lovelace += utxo['Tokens'][lovelace_unit]

        for value in values:
            transaction.outputs.append(('tx-out', f'{self.payment_address}+{value}'))
            surplus_lovelace -= value
        # This final output transaction shall contain the surplus (minus tx fee)
        transaction.outputs.append(('tx-out', f'{self.payment_address}+{surplus_lovelace}'))

        transaction.generate_draft()

        if password is not None:
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-fee
            tx_fee = transaction.calculate_min_fee()

            # Calculate the change to return the payment address
            # (minus transaction fee) and update that output respectively
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-change-to-send-back-to-payment-addr
            transaction.outputs[-1] = ('tx-out', f'{self.payment_address}+{surplus_lovelace - tx_fee}')

            transaction.submit(wallet=self, fee=tx_fee, password=password)

            # Let successful transactions be persisted to the database
            transaction.save()

        return transaction

    def mint_tokens(self, policy, quantity, to_address, spending_password, minting_password,
                    asset_name=None, metadata=dict, payment_utxo=None, change_address=None
                    ) -> AbstractTransaction:
        """
        https://docs.cardano.org/en/latest/native-tokens/getting-started-with-native-tokens.html#start-the-minting-process
        :param policy: Policy that will be used to mint the token
        :param quantity: Number of tokens to mint
        :param to_address: Address to send minted token to
        :param spending_password: Password required to decrypt wallet signing key
        :param minting_password: Password required to decrypt policy signing key
        :param asset_name: Name component of the unique asset ID (<policy_id>.<asset_name>)
        :param metadata: Metadata to include in the minting transaction
        :param payment_utxo: Specific
        :param change_address: Address to receive excess funds
        """
        surplus_address = change_address if change_address else self.payment_address

        if not payment_utxo:
            # If a payment utxo was not explicitly provided, we will use this wallet's largest
            # UTxO with the assumption that it will cover the transaction (including fees)
            sorted_lovelace_utxos = sort_utxos(self.lovelace_utxos)
            if not sorted_lovelace_utxos:
                # Let there be be at least one UTxO containing purely ADA.
                # This will be used to pay for the transaction.
                raise CardanoError(f'Inadequate funds to complete transaction')
            payment_utxo = sorted_lovelace_utxos[0]

        # Specify the asset ID and quantity of tokens to mint
        # https://docs.cardano.org/en/latest/native-tokens/getting-started-with-native-tokens.html#syntax-of-multi-asset-values
        asset_id = policy.policy_id
        if asset_name:
            asset_id = f'{asset_id}.{asset_name}'
        token_bundle = f'"{quantity} {asset_id}"'

        # Structure the token metadata according to the proposed "721" standard
        # See: https://www.reddit.com/r/CardanoDevelopers/comments/mkhlv8/nft_metadata_standard/

        transaction_model_class = get_transaction_model()
        transaction = transaction_model_class(
            tx_type=TransactionTypes.TOKEN_MINT,
            metadata=metadata,
        )
        transaction.minting_policy = policy
        transaction.minting_password = minting_password

        # Compute the amount of lovelace required to accompany this token bundle
        token_dust = CardanoUtils.min_token_dust_value(token_bundle)
        total_lovelace_being_sent = payment_utxo['Tokens'][lovelace_unit]
        lovelace_to_return = total_lovelace_being_sent - token_dust

        transaction.inputs = [('tx-in', '{}#{}'.format(payment_utxo['TxHash'], payment_utxo['TxIx']))]
        transaction.outputs = [
            ('tx-out', f'{to_address}+{token_dust}+{token_bundle}'),
            ('tx-out', f'{surplus_address}+{lovelace_to_return}')
        ]

        # Draft the transaction:
        # Produce a draft transaction in order to determine the fees required to perform the actual transaction
        # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#draft-the-transaction
        transaction.generate_draft(mint=token_bundle)

        if spending_password is not None and minting_password is not None:
            # Calculate the fee
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-fee
            tx_fee = transaction.calculate_min_fee()

            # Calculate the change to return the payment address
            # (minus transaction fee) and update that output respectively
            # https://docs.cardano.org/projects/cardano-node/en/latest/stake-pool-operations/simple_transaction.html#calculate-the-change-to-send-back-to-payment-addr
            transaction.outputs[-1] = ('tx-out', f'{surplus_address}+{lovelace_to_return - tx_fee}')

            invalid_hereafter = None
            for script in policy.script_data['scripts']:
                if script['type'] == 'before':
                    invalid_hereafter = script['slot']
                    break

            transaction.submit(
                wallet=self,
                fee=tx_fee,
                password=spending_password,
                mint=token_bundle,
                invalid_hereafter=invalid_hereafter
            )

            # Let successful transactions be persisted to the database
            transaction.save()

        return transaction


class Wallet(AbstractWallet):
    class Meta(AbstractWallet.Meta):
        swappable = 'DJANGO_CARDANO_WALLET_MODEL'



import os

from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

USER_SETTINGS = getattr(settings, 'DJANGO_CARDANO', None)

DEFAULTS = {
    'APP_DATA_PATH': os.environ.get('CARDANO_APP_DATA_PATH'),
    'CLI_PATH': os.environ.get('CARDANO_CLI_PATH'),
    'DEFAULT_TRANSACTION_TTL': 1000,
    'LOVELACE_UNIT': 'lovelace',
    'NETWORK': 'mainnet',
    'NODE_SOCKET_PATH': os.environ.get('CARDANO_NODE_SOCKET_PATH'),
    'PROTOCOL_TTL': 3600,
    # Magic numbers
    'TESTNET_MAGIC': 1097911063,
    'COIN_SIZE': 0,
    'UTXO_ENTRY_SIZE_WITHOUT_VAL': 27,
}

IMPORT_STRINGS = (

)

REMOVED_SETTINGS = (
    'SECRET_KEY',
)


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for DJANGO_CARDANO setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


class DjangoCardanoSettings:
    """
    A settings object that allows REST Framework settings to be accessed as
    properties. For example:

        from rest_framework.settings import api_settings
        print(api_settings.DEFAULT_RENDERER_CLASSES)

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.

    Note:
    This is an internal class that is only compatible with settings namespaced
    under the DJANGO_CARDANO name. It is not intended to be used by 3rd-party
    apps, and test helpers like `override_settings` may not work as expected.
    """
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'DJANGO_CARDANO', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid application setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    @staticmethod
    def __check_user_settings(user_settings):
        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError("The '{}' setting has been removed.".format(setting))
        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


django_cardano_settings = DjangoCardanoSettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)


# -----------------------------------------------------------------------------
def reload_settings(*args, **kwargs):  # pragma: no cover
    global django_cardano_settings

    setting, value = kwargs['setting'], kwargs['value']

    if setting == 'DJANGO_CARDANO':
        django_cardano_settings = DjangoCardanoSettings(value, DEFAULTS, IMPORT_STRINGS)


setting_changed.connect(reload_settings)



# Generated by Django 3.1.7 on 2021-05-26 18:02

from django.db import migrations, models
import django_cardano.fields
import django_cardano.models
import django_cardano.storage
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='MintingPolicy',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('policy_id', models.CharField(max_length=64)),
                ('name', models.CharField(blank=True, max_length=50)),
                ('script', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
                ('signing_key', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
                ('verification_key', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
            ],
            options={
                'verbose_name_plural': 'Minting Policies',
                'swappable': 'DJANGO_CARDANO_MINTING_POLICY_MODEL',
            },
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('tx_id', models.CharField(blank=True, max_length=64, null=True)),
                ('tx_file', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
                ('tx_type', models.PositiveSmallIntegerField(choices=[(1, 'Lovelace Transfer'), (2, 'Token Transfer'), (3, 'Token Mint'), (4, 'Token Consolidation'), (5, 'Lovelace Partition')])),
                ('inputs', models.JSONField(default=list)),
                ('outputs', models.JSONField(default=list)),
                ('metadata', models.JSONField(blank=True, null=True)),
            ],
            options={
                'swappable': 'DJANGO_CARDANO_TRANSACTION_MODEL',
            },
        ),
        migrations.CreateModel(
            name='Wallet',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, max_length=30)),
                ('payment_address', django_cardano.fields.CardanoAddressField(max_length=200)),
                ('payment_signing_key', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
                ('payment_verification_key', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
                ('stake_address', django_cardano.fields.CardanoAddressField(max_length=200)),
                ('stake_signing_key', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
                ('stake_verification_key', models.FileField(max_length=200, storage=django_cardano.storage.CardanoDataStorage, upload_to=django_cardano.models.file_upload_path)),
            ],
            options={
                'abstract': False,
                'swappable': 'DJANGO_CARDANO_WALLET_MODEL',
            },
            managers=[
                ('objects', django_cardano.models.WalletManager()),
            ],
        ),
    ]
